// vim: set ts=8:
#include <arpa/inet.h>

#include <ctime>
#include <iomanip>
#include <iostream>
#include <stdint.h>

#include <zlib.h>

struct Incr_ascii {
	void operator()(uint8_t buf[1024], size_t &sz) const
	{
		for (ssize_t i = sz - 1; i >= 0; i--) {
			buf[i]++;
			if (buf[i] == 0x7f) {
				buf[i] = ' ';
			} else {
				return;
			}
		}
		for (ssize_t i = sz - 1; i >= 0; i--) {
			buf[i+1] = buf[i];
		}
		buf[0] = ' ';
		sz++;
	}
};

struct Incr_binary {
	void operator()(uint8_t buf[1024], size_t &sz) const
	{
		for (ssize_t i = sz - 1; i >= 0; i--) {
			buf[i]++;
			if (buf[i] == 0x7f) {
				buf[i] = ' ';
			} else {
				return;
			}
		}
		for (ssize_t i = sz - 1; i >= 0; i--) {
			buf[i+1] = buf[i];
		}
		buf[0] = ' ';
		sz++;
	}
};

template <typename Fn>
std::string
collide(uint8_t init, const std::string &target, const std::string &mod,
    Fn incr)
{
	std::cout << std::setfill('0') << std::hex;

	size_t sz = 1;
	uint8_t buf[1024];
	buf[0] = init;

	uint32_t crc0 = crc32(0,
	    reinterpret_cast<const uint8_t *>(target.c_str()),
	    target.size());

	uint32_t crc1 = crc32(0,
	    reinterpret_cast<const uint8_t *>(mod.c_str()),
	    mod.size());

	uint32_t crc2 = 0;
	for (;;) {
		/*
	for (size_t i = 0;; i++) {
		if (i % 10485760 == 0) {
			std::cout << '\'' << std::string(buf, buf + sz)
			    << "'\n";
		}
		*/
		crc2 = crc32(crc1, buf, sz);
		if (crc2 == crc0)
			return std::string(buf, buf + sz);
		incr(buf, sz);
	}
}

int main()
{
	std::string names[] = {
		"alpha",
		//"beta",
		/*
		"gamma",
		"delta",
		"epsilon",
		"zeta",
		"eta",
		"theta",
		"iota",
		*/

		"", // end
	};

	enum { binary, ascii } type;

	//type = binary;
	type = ascii;

	std::string beta = "beta";
	for (std::string *name = names; !name->empty(); name++) {
		std::string suffix;
		struct timespec begin, end;
		clock_gettime(CLOCK_MONOTONIC, &begin);
		switch (type) {
		case binary:
			suffix = collide(0, *name, beta, Incr_binary());
			break;
		case ascii:
			suffix = collide(' ', *name, beta, Incr_ascii());
			break;
		}
		clock_gettime(CLOCK_MONOTONIC, &end);
		long long nano = (end.tv_sec - begin.tv_sec) * 1000000000LL;
		nano += end.tv_nsec - begin.tv_nsec;
		std::cout << nano/1000000000.0 << '\n';

		std::cout << "suffix: ";
		switch (type) {
		case binary:
			std::cout << std::hex << std::setfill('0');
			for (size_t i = 0; i < suffix.size(); i++)
				std::cout << std::setw(2) << suffix[i];
			std::cout << '\n';
			break;
		case ascii:
			std::cout << suffix << '\n';
			break;
		}
	}

	return 0;
}
