# -*- coding: utf-8 -*-

# Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
# Lena Olson <lena@cs.wisc.edu>
#
# You can't do anything with this but read and execute it privately.  If you
# want to do more, ask us if we'd be willing to make a more permissive
# license.

import base64
import cgi
import sys
import urllib
import zlib

from google.appengine.ext import blobstore
from google.appengine.ext import webapp
from google.appengine.ext.webapp import blobstore_handlers
from google.appengine.ext.webapp.util import run_wsgi_app

import datastore
import util

class Index(webapp.RequestHandler):
	def get(self):
		self.response.headers['Content-Type'] = (
		    'application/xhtml+xml')
		self.response.out.write(
'''<?xml version='1.0' encoding='UTF-8'?>
<!DOCTYPE html PUBLIC '-//W3C//DTD XHTML 1.0 Strict//EN'
    'http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd'>
<html xmlns='http://www.w3.org/1999/xhtml' lang='en' xml:lang='en'>
<head>
<title>MarLen CRC32 Collider</title>
<script type='text/javascript'>
//<![CDATA[

var upload_url = null;

function check_input()
{
	var form = document.getElementById('theform');

	var check = { blob: 0, data: 0, binary: 0, ascii: 0 };
	for (var name in check)
		check[name] =
		    document.getElementById('check_' + name).checked;

	var data0 = document.getElementById('data0');
	var data1 = document.getElementById('data1');
	if (check.data && /^\s*$/.test(data0.value) &&
	    !confirm('Are you sure you wanted just whitespace for the ' +
	    'target text?')) {
		return false;
	}
	if (check.data && /^\s*$/.test(data1.value) &&
	    !confirm('Are you sure you wanted just whitespace for the ' +
	    'text to modify?')) {
		return false;
	}

	if (!check.blob && !check.data) {
		alert('You must submit data.');
		return false;
	} else if (!check.binary && !check.ascii) {
		alert('You must select an append mode.');
		return false;
	}

	// use different action/enctype if a blob is being uploaded
	if (check.file) {
		// get a new url and submit when it arrives
		get_upload_url();
		return false;
	}

	form.action = '/run';
	form.enctype = 'application/x-www-form-urlencoded';
	return true;
}

// check a box
function check(check_name)
{
	document.getElementById('check_' + check_name).checked = true;
}

function start_upload(url)
{
	var form = document.getElementById('theform');
	form.action = url;
	form.enctype = 'multipart/form-data';
	form.submit();
}

function upload_url_handler()
{
	var resp = this.responseXML;
	if (this.readyState == 4 && this.status == 200) {
		if (resp) {
			var url_node = resp.getElementsByTagName('url')[0];
			start_upload(url_node.firstChild.data);
			return;
		}
		show_error('url request response was null');
	} else if (this.readyState == 4) {
		alert('HTTP error: ' + this.status + ' ' + this.statusText);
	}
}

// get a new upload URL
function get_upload_url()
{
	var client = new XMLHttpRequest();
	client.onreadystatechange = upload_url_handler;
	// append timestamp to thwart browser caching; apparently Google gets
	// to do whatever the frak they want
	var d = new Date();
	client.open('GET', '/upload_url?timestamp=' + d.getTime());
	client.send();
}

function clear_data()
{
	var client = new XMLHttpRequest();
	client.onreadystatechange = function () {}
	// append timestamp to thwart browser caching; apparently Google gets
	// to do whatever the frak they want
	var d = new Date();
	client.open('GET', '/clear?timestamp=' + d.getTime());
	client.send();
}

//]]>
</script>
</head>
<body>

<!--
    action and enctype are set by the onsubmit handler depending on file/data
    selection; action is required for XHTML-1.0-strict and enctype for gecko
-->
<form id='theform' action='' enctype='' method='post' onsubmit='return check_input()'>
<div>
    <a href='#' onclick='clear_data(); return false'>Clear Datastore and Memcache</a><br />
    <input type='radio' name='data_source' id='check_blob' />
    <label for='check_blob'>file input:</label><br />
    Target: <input type='file' name='blob0' id='blob0' onclick="check('blob');" /><br />
    To modify: <input type='file' name='blob1' id='blob1' onclick="check('blob');" /><br /><br />

    <input type='radio' name='data_source' id='check_data' />
    <label for='check_data'>data to collide with (will be converted to UTF-8 before computing the CRC):</label><br />
    Target:<br />
    <textarea id='data0' name='data0' cols='80' rows='8' onclick="check('data');" /><br />
    To modify:<br />
    <textarea id='data1' name='data1' cols='80' rows='8' onclick="check('data');" /><br /><br />

    <input type='radio' name='append_mode' id='check_binary' value='binary' />
    <label for='check_binary'>append binary data</label><br />
    <input type='radio' name='append_mode' id='check_ascii' value='ascii' />
    <label for='check_ascii'>append ASCII text</label><br />
    <input type='submit' value='Submit' />
</div>
</form>

</body>
</html>
''')

# returns an XML document with an upload URL
class Upload_url(webapp.RequestHandler):
	def get(self):
		upload_url = blobstore.create_upload_url('/upload')

		print >>sys.stderr, 'returning url `%s\'' % upload_url
		self.response.headers['Content-Type'] = 'application/xml'
		self.response.out.write('''<?xml version='1.0' encoding='UTF-8'?>
<url>%s</url>
''' % upload_url)

# receives an uploaded file, redirects to a Run.get() request
class Upload(blobstore_handlers.BlobstoreUploadHandler):
	def post(self):
		# first file: calc CRC, delete
		tgt_files = self.get_uploads('blob0')
		tgt_info = tgt_files[0]
		tgt_data = blobstore.fetch_data(tgt_info, 0,
		    tgt_info.size)
		tgt_crc = zlib.crc32(tgt_data)
		tgt_info.delete()
		if tgt_crc < 0: tgt_crc += 0x100000000

		# second file: calc CRC, store in database
		mod_files = self.get_uploads('blob1')
		mod_info = mod_files[0]
		mod_data = blobstore.fetch_data(mod_info, 0, mod_info.size)
		crc = zlib.crc32(mod_data)
		mod_info.delete()
		if mod_crc < 0: mod_crc += 0x100000000

		mode = self.request.get('append_mode')
		assert mode == 'binary' or mode == 'ascii'

		# add entry to Job table if it doesn't exist
		if not datastore.get_job(tgt_crc):
			datastore.add_job(tgt_crc, mod_crc, mode,
			    blob_info=mod_info)

		# redirect to get handler with a simple URL
		self.redirect('/run?crc=0x%08x' % tgt_crc)

class Download(webapp.RequestHandler):
	def get(self):
		# parse 'crc' parameter
		try:
			crc = int(self.request.get('crc'), 0)
		except ValueError:
			self.response.set_status(400, 'bad CRC')
			return

		# get suffix result
		job = datastore.get_job(crc)
		if job is None:
			self.response.set_status(400, 'no such job')
			return
		if job.suffix is None:
			self.response.set_status(400, 'no suffix computed')
			return

		# get data, filename, content-type
		data = job.data
		if job.append_mode == 'binary':
			content_type = 'application/octet-stream'
		elif data is None:
			info = job.blob_info
			data = blobstore.fetch_data(info, 0, info.size)
			# preserve content type of second file
			content_type = info.content_type
		else:
			content_type = 'text/plain'

		# return
		suffix = base64.b64decode(job.suffix)
		self.response.headers['Content-Type'] = content_type
		self.response.headers['Content-Disposition'] = (
		    'attachment; filename=%s' % util.filename(job))
		self.response.out.write(data + suffix)

class Run(webapp.RequestHandler):
	# receives the form input, redirects to a Run.get() request
	def post(self):
		# target data: calc CRC and discard
		tgt_data = str(self.request.get('data0').encode('utf8'))
		tgt_crc = zlib.crc32(tgt_data)
		if tgt_crc < 0: tgt_crc += 0x100000000

		# data for appending: calc CRC and store
		mod_data = str(self.request.get('data1').encode('utf8'))
		mod_crc = zlib.crc32(mod_data)
		if mod_crc < 0: mod_crc += 0x100000000

		mode = self.request.get('append_mode')
		assert mode == 'binary' or mode == 'ascii'

		# add entry to Job table if it doesn't exist
		# TODO how do we deal with choosing a master?  is there some
		# Chubby-like locking service? ... ignore -M
		if not datastore.get_job(tgt_crc):
			datastore.add_job(tgt_crc, mod_crc, mode,
			    data=mod_data)

		# redirect to get handler with a simple URL
		self.redirect('/run?crc=0x%08x' % tgt_crc)

	def get(self):
		crc_str = self.request.get('crc')

		self.response.headers['Content-Type'] = (
		    'application/xhtml+xml')
		self.response.out.write(
'''<?xml version='1.0' encoding='UTF-8'?>
<!DOCTYPE html PUBLIC '-//W3C//DTD XHTML 1.0//EN'
    'http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd'>
<html xmlns='http://www.w3.org/1999/xhtml' lang='en' xml:lang='en'>
<head>
<title>MarLen CRC32 Collider</title>
<script type='text/javascript'>
//<![CDATA[

//var RETRY_DELAY = 2500; // 2.5 s
var RETRY_DELAY = 1000; // 1 s
var FAILURE_DELAY = 10000; // 10 s

function show_result(suffix, filename, url)
{
	var status_div = document.getElementById('status');
	while (status_div.firstChild)
		status_div.removeChild(status_div.firstChild);

	var suffix_label = document.createElement('b');
	suffix_label.appendChild(document.createTextNode('Suffix: '));

	var br0 = document.createElement('br');

	var text = document.createElement('textarea');
	text.appendChild(document.createTextNode(suffix));

	var br1 = document.createElement('br');

	var link_label = document.createElement('b');
	link_label.appendChild(document.createTextNode('New file: '));

	var link = document.createElement('a');
	link.href = url;
	link.appendChild(document.createTextNode(filename));

	var nodes = [ suffix_label, br0, text, br1, link_label, link ];
	for (var i = 0; i < nodes.length; i++)
		status_div.appendChild(nodes[i]);
}

function show_error(msg)
{
	/*
	var status_div = document.getElementById('status');
	while (status_div.firstChild)
		status_div.removeChild(status_div.firstChild);

	var bold = document.createElement('b');
	bold.appendChild(document.createTextNode('Error: '));
	status_div.appendChild(bold);
	status_div.appendChild(document.createTextNode(msg));
	*/
	// asynchronous alert
	setTimeout(function () {
		alert('Error: ' + msg + ' ... will try again in 10 s.');
	}, 0);
}

// finds the first element with name 'tagname' in a document and returns its
// value
function extract_value(doc, tagname)
{
	var nodes = doc.getElementsByTagName(tagname);
	if (!nodes.length) return null;

	var node = nodes[0];
	if (!node.firstChild) return null;

	node = node.firstChild;
	if (node.data == null) return null;

	return decodeURIComponent(node.data);
}

function handler()
{
	var resp = this.responseXML;
	if (this.readyState == 4 && this.status == 200) {
		if (resp) {
			var vals = { status: 0, suffix: 0,
			    download: 0, filename: 0 };
			for (var name in vals)
				vals[name] = extract_value(resp, name);
			if (vals.status != 'completed') {
				setTimeout(wait_completion, RETRY_DELAY);
				return;
			}

			show_result(vals.suffix, vals.filename,
			    vals.download);
			return;
		}
		show_error('response is null');
	} else if (this.readyState == 4) {
		show_error(this.status + ' ' + this.statusText);
		setTimeout(wait_completion, FAILURE_DELAY);
	}
}

function wait_completion()
{
	var client = new XMLHttpRequest();
	client.onreadystatechange = handler;
	// append timestamp to thwart browser caching; apparently Google gets
	// to do whatever the frak they want
	var d = new Date();
	client.open('GET', '/check?crc=%s&timestamp=' + d.getTime());
	client.send();
}

onload = function()
{
	wait_completion();
}

//]]>
</script>
</head>
<body>

<h2>Status:</h2>
<div id='status'>Swimming...<br />
<img src='busy.gif' alt='busy' /></div>

</body>
</html>
''' % urllib.quote(crc_str))

application = webapp.WSGIApplication([
    ('/', Index),
    ('/download', Download),
    ('/run', Run),
    ('/upload', Upload),
    ('/upload_url', Upload_url),
    ], debug=True)

def main():
	run_wsgi_app(application)

if __name__ == '__main__':
	main()
