# -*- coding: utf-8 -*-

# Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
# Lena Olson <lena@cs.wisc.edu>
#
# You can't do anything with this but read  and execute it privately.  If you
# want to do more, ask us if we'd be willing to make a more permissive
# license.

import array
import base64
import time

from google.appengine.api import memcache

TASK_TIMEOUT = 45 # 45 s
TASK_EXPIRE = 5 * 60 # 5 min

def asciitize(str):
	'''Strip out non-ASCII characters.'''
	res = ''
	for c in str:
		if not (ord(c) & 0x80): res += c
	return res

def filename(job):
	if job.data is None:
		info = job.blob_info
		filename = asciitize(info.filename)
	else:
		filename = 'collision.txt'
	return filename
	
#return true iff job is already in the memcache
def check_cache(crc):
	if memcache.get(str(crc)) is not None:
		return True
	return False

def get_job_progress(crc):
	val = memcache.get(repr((crc, 'progress')))
	if val is None:
		return None
	last_end = eval(val)
	return last_end

def set_job_progress(crc, last_end):
	memcache.set(repr((crc, 'progress')), repr(last_end))

def get_task_progress(crc, end):
	'''If task timed out, return (False, None).
	If task stopped gracefully, return (False, begin).
	If task probably running, return (True, None).
	Both the 'end' argument and returned 'begin' are b64-encoded.'''
	val = memcache.get(repr((crc, end)))
	if val is None:
		return (False, None)

	(timestamp, begin, running) = eval(val)
	now = int(time.time())
	if now - timestamp < TASK_TIMEOUT or not running:
		return (False, begin)

	# otherwise it has a recent timestamp and is running
	return (True, None)

def set_task_progress(crc, end, begin, running):
	memcache.set(
	    repr((crc, end)),
	    repr((int(time.time()), begin, running)),
	    time=TASK_EXPIRE)

#add completed job to memcache (never to expire)
def add_completed(crc):
	memcache.set(str(crc), "done")

def encode_array(arr):
	'''Encode an array to base64.'''
	return base64.b64encode(arr.tostring())

def decode_array(str, type='B'):
	'''Decode an array from base64.'''
	return array.array(type, base64.b64decode(str))
